# Define FixedLengthCollector is helpful because language server protocol
# is fixed-length body protocol.  So the class can help us check out that
# if we add too much data during appending data, which can throw out error
# as soon as possible.


class FixedLengthCollector:
    """ Collector which can handle data, and automatically check out
    if we push too much data into it. """

    def __init__(self):  # type: ignore
        self.remain: int = 0
        self.data = bytearray()
        self.length_set = False

    def append(self, data: bytes) -> None:
        """ append data into collector.

        Args:
            data (bytes): data we need to append to.
        Raises:
            RuntimeError - When the length of data is more than the buffer capacity.
        """
        if self.length_set is False:
            raise RuntimeError(
                "Please call `set_length` to set the length of buffer first."
            )
        checked_length = self.remain - len(data)
        if checked_length < 0:
            raise RuntimeError("Too much data to insert into buffer.")
        self.remain -= len(data)
        self.data.extend(data)

    def set_length(self, length: int) -> None:
        """ set the length of collector.  Note that if the length is set, we can't call
        `set_length` without calling `clear` method.


        Args:
            length (int): the length of buffer
        Raises:
            RuntimeError - When the length is already set
        """
        if self.length_set:
            raise RuntimeError(
                "You have already call set_length.  This can't be called twice"
                "If you wan't to reset length, please call `clear()` first"
            )
        self.remain = length
        self.length_set = True

    def clear(self) -> None:
        """ clear the buffer. """
        self.length_set = False
        self.data.clear()
        self.remain = 0

    def __len__(self) -> int:
        """ return the length of buffer in bytes. """
        return len(self.data)

    def full(self) -> bool:
        """ return True if collect data complete. """
        return self.length_set and self.remain == 0



""" Core implementation for lsp """

import json
from json import JSONEncoder
from typing import Dict, List, Union, Type, Optional, Tuple

from ._events import (
    Close,
    EventBase,
    _HeaderEvent,
    DataReceived,
    DataSent,
    RequestReceived,
    RequestSent,
    ResponseReceived,
    MessageEnd,
    ResponseSent,
)
from ._state import IDLE, next_state, DONE, SEND_RESPONSE
from ._role import Role
from ._buffer import ReceiveBuffer
from ._collector import FixedLengthCollector
from ._errors import LspProtocolError

__all__ = ["Connection", "NEED_DATA"]


# The implementation of _SentinalCreater is very similar to _StateClassCreater
# But I think we should still split them out, because the sentinal defined here
# is not state.
class SentinalType(type):
    """ Class creater, which is use for define just a semantic const variable. """

    def __str__(self) -> str:
        return self.__name__

    def __repr__(self) -> str:
        return self.__name__


def make_sentinal(name: str) -> SentinalType:
    return SentinalType(name, (), {})


NEED_DATA = make_sentinal("NEED_DATA")


class Connection:
    """ Language server protocol Connection object.

    Args:
        role (str): represent our role.  Which can be 'cliet' or 'server'
    """

    def __init__(self, role: str):  # type: ignore
        if role == "client":
            self.our_role = Role.CLIENT
            self.their_role = Role.SERVER
        elif role == "server":
            self.our_role = Role.SERVER
            self.their_role = Role.CLIENT
        else:
            raise ValueError("The `role` value should be one of ('client', 'server')")
        self.our_state = IDLE
        self.their_state = IDLE
        self.in_buffer = ReceiveBuffer()
        self.out_collector = FixedLengthCollector()
        self.in_collector = FixedLengthCollector()

    def send(self, event: EventBase) -> bytes:
        """ send event and returns the relative bytes.  So what this function
        do is convert from event object to actual bytes, then user don't need
        to worry about bytes send to server.

        Args:
            event (EventBase): event we need to send.
        Returns:
            Bytes we can send to other side.
        """
        # transfer our state
        self.our_state = next_state(self.our_role, self.our_state, event)
        try:
            data = self._handle_event(event)
        except RuntimeError as e:
            raise LspProtocolError from e
        return data

    def _handle_event(self, event: EventBase) -> bytes:
        # convert event into bytes
        data = event.to_data()
        if isinstance(event, _HeaderEvent):
            if isinstance(event, RequestSent):
                # client fire RequestSent event, server should goto next_state according
                # to RequestReceived event
                self.their_state = next_state(
                    self.their_role, self.their_state, RequestReceived
                )
            self.out_collector.set_length(event["Content-Length"])
        elif isinstance(event, MessageEnd) and self.out_collector.remain > 0:
            raise LspProtocolError(
                f"Send Message end too quickly. Expect {len(self.out_collector)} bytes."
            )
        else:
            self.out_collector.append(data)
        return data

    def send_json(
        self, data: Union[List[Dict], Dict], encoder: Optional[Type[JSONEncoder]] = None
    ) -> bytes:
        """ helper function for sending data.

        Args:
            data (List or Dict): A valid object which can be dumps to json
            encoder (None or an subclass of json.JSONEncoder): The encoder to encode
                json, if the encoder is None, the default json.JSONEncoder will be used.
        Returns:
            Bytes that we can send to other side.
        """

        def _check_state() -> None:
            if self.our_role == Role.SERVER:
                if self.our_state is not SEND_RESPONSE or self.their_state is not DONE:
                    raise LspProtocolError(
                        "Can only send data when we receive request.\n"
                        f"But for now our_state: {self.our_state}, "
                        f"their_state: {self.their_state}"
                    )
            else:
                if self.our_state is not IDLE or self.their_state is not IDLE:
                    raise LspProtocolError(
                        "Our state or their state is not idle, may be you have send "
                        "data but havn't called `go_next_circle` to refresh state?"
                    )

        def _set_state() -> None:
            self.our_state = DONE
            if self.our_role == Role.CLIENT:
                self.their_state = SEND_RESPONSE

        # do state checking.
        _check_state()
        binary_data = json.dumps(data, cls=encoder).encode("utf-8")
        header_event = _HeaderEvent({"Content-Length": len(binary_data)})
        _set_state()
        return header_event.to_data() + binary_data

    def next_event(self) -> Union[SentinalType, EventBase]:
        """ Parse the next event out of incoming buffer, and return it.

        Note that this method will change the connection according to the status of
        incoming buffer.

        Returns:
            It will return one of the following results:
            1. An event object to indicate that what happened to our incoming buffer.
            2. A special constant NEED_DATA, which indicate that user need to receive
            data from remote server, and calling receive(data).
        """
        if self.our_role is Role.CLIENT and self.our_state is not DONE:
            raise LspProtocolError("Client can only accept data after it send request.")
        event = self._extract_event()
        if isinstance(event, EventBase):
            their_event: Union[type, EventBase]
            # when we get RequestReceived/ResponseReceived/DataReceived event, we
            # should change other side's state by RequestSent/ResponseSent/DataSent
            # event.
            if isinstance(event, _HeaderEvent):
                if isinstance(event, RequestReceived):
                    # when server get RequestReceived event, it should change
                    # the state according to this event.
                    self.our_state = next_state(self.our_role, self.our_state, event)
                    their_event = RequestSent
                elif isinstance(event, ResponseReceived):
                    their_event = ResponseSent
            elif isinstance(event, DataReceived):
                their_event = DataSent
            else:
                their_event = event.__class__
            # transfer their_state
            self.their_state = next_state(
                self.their_role, self.their_state, their_event
            )
        return event

    def receive(self, data: bytes) -> None:
        """ Receive data and feed it to our incoming buffer.  Then we can call
        `next_event` to extrace out incoming events.

        Args:
            data (bytes): the data we received.
        """
        self.in_buffer.append(data)

    def _extract_event(self) -> Union[SentinalType, EventBase]:
        """ parse and extract event from incoming buffer. """
        # we don't get any header yet.
        if self.in_buffer.header_bytes is None:
            header = self.in_buffer.try_extract_header()
            if header is None:  # header data doesn't completely received.
                return NEED_DATA
            else:
                event_obj: _HeaderEvent
                if self.our_role == Role.SERVER:
                    event_obj = RequestReceived(header)
                else:
                    event_obj = ResponseReceived(header)
                self.in_collector.set_length(int(event_obj["Content-Length"]))
                return event_obj
        else:
            data = self.in_buffer.try_extract_data()
            if data is None:
                if self.in_collector.remain == 0:
                    return MessageEnd()
                return NEED_DATA
            else:
                self.in_collector.append(data)
                return DataReceived({"data": data})

    def go_next_circle(self) -> None:
        """ go to next request/response circle.

        Raises:
            LspProtocolError - When our state and their state is not done yet.
        """
        # As server: when we receive Notification complete
        # As client: when we cend Notification complete
        # We can just go_next_circle.  But just ensure that client's state is DONE
        # server's state is SEND_RESPONSE
        if self.our_role is Role.CLIENT:
            if self.our_state is not DONE or (
                self.their_state not in (SEND_RESPONSE, DONE)
            ):
                raise LspProtocolError(
                    "As client, my state is not done or server's state is invalid"
                )
        else:
            if (
                self.our_state not in (SEND_RESPONSE, DONE)
                or self.their_state is not DONE
            ):
                raise LspProtocolError(
                    "As client, my state is not SEND_RESPONSE or DONE, "
                    "or client's state is invalid"
                )
        self.our_state = IDLE
        self.their_state = IDLE
        self.in_buffer.clear()
        self.out_collector.clear()
        self.in_collector.clear()

    def get_received_data(
        self, raw: bool = False
    ) -> Tuple[Dict, Union[bytes, Dict, List]]:
        """ A helper method to extract our received data.  This method is useful
        when we get `MessageEnd` event(which indicate we have received data completely).
        And it returns a eaiily-handled python objects.

        Args:
            raw (bool): Indicate that if we should return raw data(in bytes).  Or
                deserialized data.

        Returns:
            A tuple contains (header, data), header has type dict.
            And data's type can be bytes when raw is True(So it can be serialize by
            other json library on user code).  Or it can be a dict or list object.

        Raises:
            Raise RuntimeError when we don't receive data completely.
        """
        header = self.in_buffer.try_extract_header()
        if header is None or not self.in_collector.full():
            raise RuntimeError(
                "Receive data incompletely.  Please call `next_event()` until"
                "Received MessageEnd event"
            )
        if raw is False:
            return header, json.loads(self.in_buffer.raw)
        else:
            return header, bytes(self.in_buffer.raw)

    def close(self) -> None:
        """ Close the connection, make both states go to closed. """
        self.our_state = next_state(self.our_role, self.our_state, Close)
        self.their_state = next_state(self.their_role, self.their_state, Close)



from typing import Optional, Dict


class ReceiveBuffer:
    """ Inner data buffer.  It can receive data, and extract our header part and body
    part later. """

    def __init__(self):  # type: ignore
        self.raw = bytearray()
        self.body_pointer: int = 0
        self._header_bytes: Optional[bytearray] = None
        self.header: Optional[Dict[str, str]] = None

    @property
    def header_bytes(self) -> Optional[bytearray]:
        return self._header_bytes

    @header_bytes.setter
    def header_bytes(self, value: Optional[bytearray]) -> None:
        def _extract_header() -> Dict[str, str]:
            header_str = self._header_bytes.decode("ascii")  # type: ignore
            results = {}
            rows = header_str.split("\r\n")
            for row in rows:
                key, val = row.split(": ")
                results[key] = val
            return results

        self._header_bytes = value
        if value is None:
            self.header = None
        else:
            self.header = _extract_header()

    def append(self, data: bytes) -> None:
        """ Append data into buffer.

        Args:
            data (bytes): the data we need to append.
        """
        self.raw.extend(data)

    def try_extract_header(self) -> Optional[Dict[str, str]]:
        """ Try to extract the header part in the buffer.

        Returns:
            When the buffer received completely header data, then return
            A dict.  Else we return None.
        """

        if self.header is not None:
            return self.header
        data_splitted = self.raw.split(b"\r\n\r\n")
        if len(data_splitted) == 1:  # so we don't receive header data complete yet.
            return None
        else:
            # we have receive header completely, so we can extract header, and if there
            # are any data inputed, we save it to the raw, which indicate that it's
            # un-handled.
            self.header_bytes, self.raw = (
                bytearray(data_splitted[0]),
                bytearray(data_splitted[1]),
            )
            return self.header

    def try_extract_data(self) -> Optional[bytes]:
        """ Try to extract the actual data in buffer.  Note that we should call
        `try_extract_header` first to extract header out.

        Returns:
            When there are data in the buffer, return it.  Return None to indicate
            there are no data in the buffer.

        Raises:
            RuntimeError - When the buffer doesn't completely handle header data.
        """
        if self.header_bytes is None:
            raise RuntimeError(
                "Header is un-handled yet, please call `try_extract_header` first."
            )
        # TODO: need to rewrite the implementation.  Because the slice operation will
        # copy memeory, and it may be high cost.
        if self.body_pointer == len(self.raw):
            return None
        # fmt: off
        data = self.raw[self.body_pointer:]
        # fmt: on
        self.body_pointer += len(data)
        return data

    def clear(self) -> None:
        """ clear the buffer.  Which is useful when Connection want
        to start the next circle. """
        self.header_bytes = None
        self.raw.clear()
        self.body_pointer = 0



class LspProtocolError(BaseException):
    """ exception for lsp protocol error.  Mainly contains the following
    errors:
    1. invalid connection state transfer.
    2. send MessageEnd event, but the length of sending buffer doesn't match
    the content of length which defined in the header.
    """

    pass



""" Contains events definition which is exposed by lsp.

We can initialize EventClass by this way:
event = RequestSent({'Content-Length': 90})
"""

import warnings
import json
from typing import Set, List, Tuple, Dict, Any, Optional


__all__ = [
    # Mainly used by server
    "RequestReceived",
    "ResponseSent",
    # Mainly userd by client
    "ResponseReceived",
    "RequestSent",
    # Common
    "DataSent",
    "DataReceived",
    "MessageEnd",
    "Close",
    "MessageEnd",
]


# Details:
# 1. MetaClass should derived from type
# 2. When we are creating a class, the MetaClass' __new__, __init__ function
#    will be invoked one by one, just like type(class_name, bases, attributes)
# For more information, please see two useful links:
# https://stackoverflow.com/questions/100003/what-are-metaclasses-in-python
# https://docs.python.org/3/reference/datamodel.html#metaclasses

# NOTE: because it's just the helper implementation of event class definition
# so we needn't test for it.
class _EventBaseMeta(type):  # pragma: no cover
    """ MetaClass for lsp events.

    When a class using this metaclass, the '_required' fields will be
    auto-generated.  And it required class should have '_fields' attribute.
    """

    def __new__(cls, cls_name: str, bases: Tuple, attrs: Dict) -> type:
        if bases is None:
            assert (
                "_fields" in attrs
            ), "The new creating class must contains '_fields' attribute"
        return super(_EventBaseMeta, cls).__new__(cls, cls_name, bases, attrs)

    def __init__(cls, cls_name, bases, attrs):  # type: ignore
        if not hasattr(cls, "_defaults"):
            cls._defaults: List[Tuple[str, str]] = []

        # check if there are too much optional_fields.
        # Any fields which is in _defaults but not in _fields is not allowed.
        optional_fields = set([default_field[0] for default_field in cls._defaults])
        invliad = optional_fields - cls._fields
        if invliad:
            raise ValueError(
                f"Optional contains fields {invliad} which is not existed on _fields"
            )

        cls._required = cls._fields - optional_fields


class EventBase(metaclass=_EventBaseMeta):
    """ Very very base definition for lsp events.

    Attributes:
        _fields (Set[str]): A set of fields in the events.
        _defaults (List[Tuple[str, Any]]): A list of tuple contains
            (field_name, default_value).
        _required (Set[str]): It's worth to know that the _required fields will
            be FILLED AUTOMATICALLY during the class creation.  Subclass can use
            this field WITHOUT worry about creation.  When a fields is not optional
            (which lays in _defaults), it's required.

    Example:
        class SendHeader(EventBase):
            _fields = {"field1", "field2"}
            _defaults = [("field1", "go")]

        After the class creation, the value of SendHeader._required is {"field2"}
    """

    _fields: Set[str] = set()
    _defaults: List[Tuple[str, Any]] = []

    def __init__(self, kwargs: Optional[Dict] = None):  # type: ignore
        if kwargs is None:
            return
        keys = set(kwargs.keys())
        missing_required = self.__class__._required - keys
        # check for fields
        if missing_required:
            raise ValueError(f"Missing required fields: {missing_required}")
        too_much_fields = keys - self._fields
        if too_much_fields:
            warnings.warn(
                f"There are too much fields: {too_much_fields}, I will ignore them."
            )

        self.__dict__.update(self._defaults)
        self.__dict__.update(kwargs)

    def __getitem__(self, key: str) -> Any:
        return getattr(self, key)

    def to_data(self, encoding: str = "ascii") -> bytes:
        """ convert event into bytes.

        Args:
            encoding (str): the encoding we use to encode data.
        Returns:
            bytes which represent event.
        """
        raise NotImplementedError()


class DataEvent(EventBase):
    """ The DataEvent are fired when we deal with actual data. """

    _fields = {"data"}

    def to_data(self, encoding: str = "utf-8") -> bytes:
        if isinstance(self["data"], (bytes, bytearray)):
            data = bytes(self["data"])
        elif isinstance(self["data"], str):
            data = self["data"].encode(encoding)
        elif isinstance(self["data"], (list, dict)):
            data = json.dumps(self["data"]).encode(encoding)
        return data


class DataReceived(DataEvent):
    """ The DataReceived events are fired when we get request data. """

    _fields = {"data"}


class DataSent(DataEvent):
    """ The DataSent events are fired when we send request/response data. """

    _fields = {"data"}


class _HeaderEvent(EventBase):
    """ Fired when header is sent. """

    _fields = {"Content-Length", "Content-Type"}
    _defaults = [("Content-Type", "application/vscode-jsonrpc; charset=utf-8")]

    def to_data(self, encoding: str = "ascii") -> bytes:
        row_spliter = "\r\n"
        rows = []
        for field in self._fields:
            rows.append(f"{field}: {self[field]}")
        # because the splitter between header part and
        # data part is \r\n, we need to add two \r\n
        rows.append("")
        rows.append("")
        # The header is encoded in ascii, by the definition of lsp
        return row_spliter.join(rows).encode("ascii")


class RequestReceived(_HeaderEvent):
    """ The RequestReceived events are fired when we get request header. """

    pass


class RequestSent(_HeaderEvent):
    """ Fired when request header is sent. """

    pass


class ResponseReceived(_HeaderEvent):
    """ The ResponseReceived events are fired when we get response header. """

    pass


class ResponseSent(_HeaderEvent):
    """ Fired when response header is sent. """

    pass


class Close(EventBase):
    """ Fired when we need to close connection. """

    def to_data(self, encoding: str = "ascii") -> bytes:
        return b""


class MessageEnd(EventBase):
    """ Fired when we send data complete. """

    def to_data(self, encoding: str = "ascii") -> bytes:
        return b""



""" Define lsp connection state for client side and server side. """

import textwrap

from typing import Type, Dict, Union
from ._role import Role
from ._events import (
    RequestSent,
    DataSent,
    MessageEnd,
    Close,
    RequestReceived,
    ResponseSent,
    EventBase,
)
from ._errors import LspProtocolError

__all__ = ["IDLE", "SEND_BODY", "SEND_RESPONSE", "DONE", "CLOSED"]


class _StateClassCreater(type):
    """ Creater for state class. """

    def __str__(self) -> str:
        return f"<State: {self.__name__}>"

    def __repr__(self) -> str:
        return self.__name__


def make_state(state_name: str) -> Type:
    """ make state as a class.

    Args:
        state_name (str): the state name.
    Returns:
        A state relative class
    """
    return _StateClassCreater(state_name, (), {})


def next_state(role: Role, current_state: Type, event: Union[type, EventBase]) -> Type:
    """ given the role with current state, find the next state when received
    the given event.

    Args:
        role (_role.Role): The role of connection.
        current_state (type): the current state of connection.
        event (EventBase or type of EventBase): The event we received.
    Returns:
        An instance of type indicate the next state.
    Raises:
        LspProtocolError - if the current_state is not a valid state of role.
            Or we can't find next state
    """
    event_cls = event.__class__ if isinstance(event, EventBase) else event

    state_machine = _client_state if role == Role.CLIENT else _server_state
    if current_state not in state_machine:
        raise LspProtocolError(f"The given state {repr(current_state)} is invalid.")
    next_state = state_machine[current_state].get(event_cls, None)
    if not next_state:
        raise LspProtocolError(
            textwrap.indent(
                f"\nThe event is invalid.  More information: "
                f"role - {role}; state - {current_state}; event - {event_cls}",
                " " * 4,
            )
        )
    return next_state


# States definition
IDLE = make_state("IDLE")
SEND_BODY = make_state("SEND_BODY")
DONE = make_state("DONE")
CLOSED = make_state("CLOSED")
SEND_RESPONSE = make_state("SEND_RESPONSE")


# state machine definieion
_client_state: Dict[type, Dict] = {
    IDLE: {RequestSent: SEND_BODY, Close: CLOSED},
    SEND_BODY: {DataSent: SEND_BODY, Close: CLOSED, MessageEnd: DONE},
    DONE: {Close: CLOSED},
    CLOSED: {},
}

_server_state: Dict[type, Dict] = {
    IDLE: {RequestReceived: SEND_RESPONSE, Close: CLOSED},
    SEND_RESPONSE: {ResponseSent: SEND_BODY, Close: CLOSED},
    SEND_BODY: {DataSent: SEND_BODY, Close: CLOSED, MessageEnd: DONE},
    DONE: {Close: CLOSED},
    CLOSED: {},
}



from setuptools import setup, find_packages


with open("README.rst", "r") as f:
    DESCRIPTION = f.read()


# read out version.
with open("lsp/_version.py", "r") as f:
    exec(f.read())


setup(
    name="lsp",
    version=__version__,
    author="WindSoilder",
    author_email="WindSoilder@outlook.com",
    url="https://github.com/WindSoilder/lsp",
    description="Sans-IO pattern, language server protocol implementation",
    long_description=DESCRIPTION,
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: Implementation :: CPython",
        "Intended Audience :: Developers",
    ],
    packages=find_packages(".", exclude=("*tests",)),
    python_requires=">=3.6",
    license="MIT",
    zip_safe=False,
)



__all__ = ["LspProtocolError"]


from ._errors import LspProtocolError
from ._connection import Connection, NEED_DATA
from ._events import (
    # Mainly used by server
    RequestReceived,
    ResponseSent,
    # Mainly userd by client
    ResponseReceived,
    RequestSent,
    # Common
    DataSent,
    DataReceived,
    MessageEnd,
    Close,
    MessageEnd,
)
from ._state import IDLE, SEND_BODY, SEND_RESPONSE, DONE, CLOSED
from ._version import __version__

__all__ += _connection.__all__
__all__ += _events.__all__
__all__ += _state.__all__
__all__ += [__version__]
